/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

#include <string.h>
#include <glib.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <libhal.h>
#include <NetworkManager.h>
#include <nm-connection.h>
#include <nm-setting-ip4-config.h>

#include "parser.h"
#include "writer.h"

#define CONF_FILE SYSCONFDIR "/NetworkManager/nm-system-settings.conf"
#define CONF_DIR  SYSCONFDIR "/sysconfig/network"

/* Connections */

static NMDeviceType
get_device_type (const char *iface)
{
    DBusGConnection *bus;
    LibHalContext *hal_ctx;
    DBusError error;
    NMDeviceType type = NM_DEVICE_TYPE_UNKNOWN;

    bus = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);
    if (!bus) {
        g_warning ("Could not connect to system bus.");
        return type;
    }

    hal_ctx = libhal_ctx_new ();
    libhal_ctx_set_dbus_connection (hal_ctx, dbus_g_connection_get_connection (bus));

    dbus_error_init (&error);
    if (libhal_ctx_init (hal_ctx, &error)) {
        char **devices = NULL;
        int num_devices = 0;

        devices = libhal_manager_find_device_string_match (hal_ctx, "net.interface",
                                                           iface, &num_devices, &error);

        if (num_devices == 1) {
            char *type_str;

            type_str = libhal_device_get_property_string (hal_ctx, devices[0], "info.category", NULL);
            if (type_str) {
                if (!strcmp (type_str, "net.80203"))
                    type = NM_DEVICE_TYPE_ETHERNET;
                else if (!strcmp (type_str, "net.80211"))
                    type = NM_DEVICE_TYPE_WIFI;
                else
                    g_warning ("Unsupported device type '%s'", type_str);

                libhal_free_string (type_str);
            }
        } else
            g_warning ("Device '%s' not found", iface);

        if (devices)
            libhal_free_string_array (devices);
    } else
        g_warning ("HAL initialization failed: %s", error.message);

    if (dbus_error_is_set (&error))
        dbus_error_free (&error);

    libhal_ctx_shutdown (hal_ctx, NULL);
    libhal_ctx_free (hal_ctx);

    dbus_g_connection_unref (bus);

    return type;
}

static NMConnection *
read_connection (const char *iface, NMDeviceType type)
{
    NMConnection *connection;

    connection = parse_ifcfg (iface, type);
    if (connection) {
        NMSettingIP4Config *ip4;

        ip4 = (NMSettingIP4Config *) nm_connection_get_setting (connection, NM_TYPE_SETTING_IP4_CONFIG);
        if (ip4 && !strcmp (nm_setting_ip4_config_get_method (ip4), NM_SETTING_IP4_CONFIG_METHOD_MANUAL)) {
            /* static IP connection, merge the default route and DNS information */

            if (nm_setting_ip4_config_get_num_addresses (ip4) > 0)
                nm_ip4_address_set_gateway (nm_setting_ip4_config_get_address (ip4, 0),
                                            parser_parse_routes (CONF_DIR "/routes"));

            parser_parse_resolv_conf (ip4, SYSCONFDIR "/resolv.conf");
        }
    }

    return connection;
}

static void
connections_merge (void)
{
    GDir *dir;

    dir = g_dir_open (CONF_DIR, 0, NULL);
    if (dir) {
        const char *f;

        while ((f = g_dir_read_name (dir)) != NULL) {
            if (g_str_has_prefix (f, "ifcfg-") && 
                strcmp (f, "ifcfg-lo") &&
                !g_str_has_suffix (f, "~") &&
                !g_str_has_suffix (f, ".bak")) {

                const char *iface = f + 6;
                NMDeviceType type = get_device_type (iface);

                if (type != NM_DEVICE_TYPE_UNKNOWN) {
                    NMConnection *connection;

                    g_print ("Converting '%s' (interface: %s: type: %d)\n", f, iface, type);

                    connection = read_connection (iface, type);
                    if (connection) {
                        char *path = NULL;
                        GError *error = NULL;

                        write_connection (connection, &path, &error);
                        if (error) {
                            g_warning ("%s", error->message);
                            g_error_free (error);
                        }

                        g_free (path);
                        g_object_unref (connection);
                    }
                }
            }
        }

        g_dir_close (dir);
    } else
        g_warning ("Could not open " CONF_DIR " directory");
}

/* */

int
main (int argc, char *argv[])
{
    int i;

    g_type_init ();
    g_set_prgname (argv[0]);

    for (i = 1; i < argc; i++) {
        const char *a = argv[i];

        if (a && !strcmp (a, "--connections"))
            connections_merge ();
    }

    return 0;
}
